""" Extensions to Matplotlib: Cursor class, and ginput function.

This module contains a Cursor class (cursor that moves with the mouse) and
a ginput function similar to the matplotlib ginput, but with a cursor
and allowing zooming/panning.
"""

# TODO - To allow cursor to appear on multiple figures, it is necessary to
# connect figure events to callbacks for every existing figure --> do a figure list
# and connect one by one.

# TODO -- add option to pick exact already drawn datapoints close to the click


import time

from .interactive_object import InteractiveObject


class Cursor(InteractiveObject):
    """Cursor following the mouse on any axes of a matplotlib figure.

    Interactive cursor appearance:
    - space bar to toggle visibility (on/off)
    - shift + up/down arrows: increase or decrease width (linewidth)
    - shift + left/right arrows: cycle through different cursor colors

    Interactively extracting figure data:
    - 'a' or left click to add point
    - 'z' or right click to remove point
    - 'enter' or middle click for stopping

    Parameters
    ----------
    All parameters optional so that a cursor can be created by `Cursor()`.

    - `fig` (matplotlib figure, default: current figure, specified as None).
    - `color` (matplotlib's color, default: None (class default value)).
    - `c` (shortcut for `color`)
    - `linestyle` (matplotlib's linestyle, default: dotted ':').
    - `linewidth` (float, default: 1.0). Line width.
    - `horizontal`: if True (default), display horizontal line of cursor
    - `vertical`: if True (default), display vertical line of cursor
    - `blit` (bool, default: True). Blitting for performance.
    - `show_clicks` (bool, default:False). Mark location of clicks.
    - `record_clicks` (bool, default False). Create a list of click positions.

    The 3 following parameters can be 1, 2, 3 (left, middle, right mouse btns).
    - `mouse_add` (int, default 1). Adds a (x, y) point by clicking.
    - `mouse_pop` (int, default 3). Removes most recently added point.
    - `mouse_stop`(int, default 2). Stops click recording. Same as reaching n.

    The 3 following parameters are useful for ginput-like functions.
    - `n` (int, default 1000). Cursor deactivates after n clicks.
    - `block` (bool, default False). Block console until nclicks is reached.
    - `timeout` (float, default 0, i.e. infinite) timeout for blocking.

    The last 2 parameters customize appearance of click marks when shown.
    - `mark_symbol` (matplolib's symbol, default: '+')
    - `mark_size` (matplotlib's markersize, default: 10)

    Useful class methods
    --------------------
    - `erase_marks()`: erase click marks on the plot.
    - `erase_data()`: reset recorded click data.

    Useful class attributes
    -----------------------
    - `fig`: matplotlib figure the cursor is active in. Fixed.
    - `ax`: matplotlib axes the cursor is active in. Changes in subplots.
    - `visible`: bool, sets whether cursor drawn or not when in axes.
    - `inaxes`: book, true when mouse (and thus cursor) is in axes
    - `clicknumber`: track the number of recorded clicks.
    - `clickdata`: stores the (x, y) data of clicks in a list.
    - `marks`: list of matplotlib artists containing all click marks drawn.
    """

    name = 'Cursor'

    def __init__(self, fig=None, color=None, c=None, linestyle=':', linewidth=1,
                 horizontal=True, vertical=True,
                 blit=True, show_clicks=False, record_clicks=False,
                 mouse_add=1, mouse_pop=3, mouse_stop=2,
                 n=1000, block=False, timeout=0,
                 mark_symbol='+', mark_size=10):
        """Note: cursor drawn only when the mouse enters axes."""

        super().__init__(fig, color=color, c=c, blit=blit, block=block)

        # Cursor state attributes
        self.press = False  # active when mouse is currently pressed
        self.visible = True  # can be True even if cursor not drawn (e.g. because mouse is outside of axes)
        self.inaxes = False  # True when mouse is in axes

        # Appearance options
        self.style = linestyle
        self.width = linewidth
        self.marksymbol = mark_symbol
        self.marksize = mark_size
        self.horizontal = horizontal
        self.vertical = vertical

        # Recording click options
        self.markclicks = show_clicks
        self.recordclicks = record_clicks
        self.clickbutton = mouse_add
        self.removebutton = mouse_pop
        self.stopbutton = mouse_stop

        # Recording click data
        self.clicknumber = 0  # tracks the number of clicks
        self.n = n  # maximum number of clicks, after which cursor is deactivated
        self.clickdata = []  # stores the (x, y) data of clicks in a list
        self.marks = []  # list containing all artists drawn

        self.fig.canvas.draw()

        # the blocking option below needs to be after connect()
        if self.block:
            self.fig.canvas.start_event_loop(timeout=timeout)

    def __repr__(self):

        name = self.__class__.name
        base_message = f'{name} on Fig. {self.fig.number}.'

        if self.clickbutton == 1:
            button = "left"
        elif self.clickbutton == 2:
            button = 'middle'
        elif self.clickbutton == 3:
            button = 'right'
        else:
            button = 'unknown'

        if self.markclicks:
            add_message = f"Leaves '{self.marksymbol}' marks when {button} mouse button is pressed. "
        else:
            add_message = ''

        if self.recordclicks:
            add_message += f'Positions of clicks with the {button} button is recorded in the clickdata attribute.'
        else:
            add_message += 'Positions of clicks not recorded.'

        return base_message + ' ' + add_message

# =========================== main cursor methods ============================

    def create(self, event):
        """Draw a cursor (h+v lines) that stop at the edge of the axes."""
        ax = self.ax
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()

        self.delete_others('fig')  # delete all other existing cursors on the figure

        x, y = event.xdata, event.ydata

        # horizontal and vertical cursor lines, the animated option is for blitting

        self.cursor_lines = {}

        if self.horizontal:
            hline, = ax.plot([xmin, xmax], [y, y], color=self.color,
                             linewidth=self.width, linestyle=self.style,
                             animated=self.__class__.blit)
            self.cursor_lines['horizontal'] = hline

        if self.vertical:
            vline, = ax.plot([x, x], [ymin, ymax], color=self.color,
                             linewidth=self.width, linestyle=self.style,
                             animated=self.__class__.blit)
            self.cursor_lines['vertical'] = vline

        # because plotting the lines can change the initial xlim, ylim
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        self.all_artists = tuple(self.cursor_lines.values())
        # Note: addition to all_objects is made automatically by InteractiveObject parent class
        self.__class__.moving_objects.add(self)

        # Below is for cursor to be visible upon creation
        if self.__class__.blit:
            self.fig.canvas.blit(self.ax.bbox)
        else:
            self.fig.canvas.draw()

    def update_position(self, event):
        """Update position of the cursor to follow mouse event."""

        x = event.xdata  # For cursors it is sufficient to work with data coordinates
        y = event.ydata  # (no need to go to pixels as the cursor is always in axes)

        # accommodates changes in axes limits while cursor is on
        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()

        if self.horizontal:
            hline = self.cursor_lines['horizontal']
            hline.set_xdata([xmin, xmax])
            hline.set_ydata([y, y])

        if self.vertical:
            vline = self.cursor_lines['vertical']
            vline.set_xdata([x, x])
            vline.set_ydata([ymin, ymax])

    def set_press_info(self, event):
        self.press_info = {'currently_pressed': True,
                           'click_position': (event.xdata, event.ydata)}

    def erase_marks(self):
        """Erase plotted clicks (marks) without removing click data"""
        for mark in self.marks:
            mark.remove()
        self.fig.canvas.draw()

    def erase_data(self):
        """Erase data of recorded clicks"""
        self.clickdata = []

    def add_point(self, pos):
        """Add point to the click data (triggered by click or key press)"""
        x, y = pos
        if self.recordclicks:
            self.clickdata.append((x, y))
            self.clicknumber += 1

        if self.markclicks:
            mark, = self.ax.plot(x, y, marker=self.marksymbol, color=self.color,
                                 markersize=self.marksize)
            self.marks.append(mark)
        self.fig.canvas.draw()

    def remove_point(self):
        """Add point to the click data (triggered by click or key press)"""
        if self.recordclicks:
            if self.clicknumber == 0:
                pass
            else:
                self.clicknumber -= 1
                self.clickdata.pop(-1)  # remove last element

        if self.markclicks:
            if len(self.marks) == 0:
                pass
            else:
                mark = self.marks.pop(-1)
                mark.remove()
        self.fig.canvas.draw()

# ============================= callback methods =============================

    def on_enter_axes(self, event):
        """Create a cursor when mouse enters axes."""
        self.inaxes = True
        self.ax = event.inaxes
        if self.visible:
            self.create(event)
        if self.__class__.blit:
            self.__class__.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)

    def on_leave_axes(self, event):
        """Erase cursor when mouse leaves axes."""
        self.inaxes = False
        if self.visible and not self.press_info['currently_pressed']:
            self.erase()

    def on_motion(self, event):
        """Update position of the cursor when mouse is in motion."""
        # do nothing if pressed to avoid weird interactions with panning
        if self.visible and self.inaxes and not self.press_info['currently_pressed']:
            self.update_graph(event)

    def on_mouse_press(self, event):
        """If mouse is pressed, deactivate cursor temporarily."""
        self.set_press_info(event)
        if self.visible and self.inaxes:
            self.erase()

    def on_mouse_release(self, event):
        """When releasing click, reactivate cursor and redraw figure.

        This is in order to accommodate potential zooming/panning.
        """
        self.press_info['currently_pressed'] = False
        if self.visible and self.inaxes:
            self.create(event)

        # I don't understand why I need to do the hack below to not have a
        # strange re-appearance of the background before zooming (panning is ok)
        # when the mouse go into motion again (not rightaway). Even more
        # surprising is that if I shortcut on_motion by calling update_graph
        # directly here, it does not work.
        self.__class__.initiating_motion = True  # to reactivate cursor

        # See if click needs to be recorded.

        x, y = (event.xdata, event.ydata)
        # line below avoids recording clicks during panning/zooming
        if (x, y) == self.press_info['click_position']:

            if event.button == self.clickbutton:
                self.add_point((x, y))

            elif event.button == self.removebutton:
                self.remove_point()

        if self.clicknumber == self.n or event.button == self.stopbutton:
            print('Cursor disconnected (max number of clicks, or stop button pressed).')
            self.delete()

    def on_key_press(self, event):
        """Key press controls. Space bar toggles cursor visibility.

        All controls:
            - space bar: toggles cursor visibility
            - up/down arrows: increase or decrease cursor size
            - left/right arrows: cycles through colors
            - "a" : add point
            - "z" : cancel last point
            - enter : stop recording
        """
# ----------------- changes in appearance of cursor --------------------------

        commands_color = ['shift+right', 'shift+left']  # keys to change color
        commands_width = ['shift+up', 'shift+down']

        if event.key == " ":  # Space Bar
            if self.inaxes:  # create or delete cursor only if it's in axes
                self.erase() if self.visible else self.create(event)
            self.visible = not self.visible  # always change visibility status

        if event.key == commands_width[0]:
            self.width += 0.5

        if event.key == commands_width[1]:
            self.width = self.width - 0.5 if self.width > 0.5 else 0.5

        if event.key in commands_color:
            # finds at which position the current color is in the list
            colorindex = self.__class__.colors.index(self.color)
            if event.key == commands_color[1]:
                colorindex -= 1
            else:
                colorindex += 1
            colorindex = colorindex % len(self.__class__.colors)
            self.color = self.__class__.colors[colorindex]

        if event.key in commands_color + commands_width:
            self.erase()  # easy way to not have to update artist
            self.create(event)

# ------------------- recording or removing click data -----------------------

        x, y = (event.xdata, event.ydata)

        if event.key == 'a':
            self.add_point((x, y))

        # I use 'z' here because backspace (as used in ginput) interferes
        # with the interactive "back" option in matplotlib
        elif event.key == 'z':
            self.remove_point()

# --------------------implement changes on graph -----------------------------

        # hack to see changes directly and to prevent display bugs
        self.__class__.initiating_motion = True
        self.update_graph(event)

# ------------------------ stop if necessary ---------------------------------

        if self.clicknumber == self.n or event.key == 'enter':
            print('Cursor disconnected (max number of clicks, or stop button pressed).')
            self.delete()

    def on_pick(self, event):
        """Contrary to draggble objects, no self-picking here."""
        pass


# =========================== ginput-like function ==========================


def ginput(n=1, timeout=0, show_clicks=True,
           mouse_add=1, mouse_pop=3, mouse_stop=2,
           color=None, c=None, linestyle=':', linewidth=1,
           horizontal=True, vertical=True,
           blit=True):
    """Improved ginput function (graphical data input) compared to Matplotlib's.

    In particular, a cursor helps for precise clicking and zooming/panning
    do not add extra click data.

    Key shortcuts and mouse clicks follow the Cursor class behavior,
    in particular the key shortcuts are `a`, `z`, `enter` instead of
    any key, backspace and enter. See Cursor class documentation for more info.
    All Cursor class interactive features are usable.

    Parameters
    ----------
    Parameters are exactly the same as matplotlib.pyplot.ginput, see
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.ginput.html
    with the following additional parameters (see drapo.Cursor):

    - color / c (default None, i.e. auto)
    - linestyle (default ':')
    - linewidth (default 1)
    - horizontal (default True)
    - vertical (default True)
    - blit (bool, default True)

    Returns
    -------
    List of tuples corresponding to the list of clicked (x, y) coordinates.
    """
    c = Cursor(block=True, record_clicks=True, show_clicks=show_clicks, n=n,
               color=color, c=c, linestyle=linestyle, linewidth=linewidth,
               horizontal=horizontal, vertical=vertical,
               mouse_add=mouse_add, mouse_stop=mouse_stop, mouse_pop=mouse_pop,
               blit=blit)
    data = c.clickdata
    time.sleep(0.2)  # just to have time to see the last click and its mark
    c.erase_marks()
    return data



"""Module to be able to select active figure and axes by clicking on them."""

# TODO -- add on_close behavior (e.g. disconnect, but only on the closed fig)
# TODO -- similarly, disconnect when the last figure is closed ? I don't know if this is useful
# TODO -- use key shortcuts to validate / cancel
# TODO -- add function that returns the selected ax/fig (probably needs block)


import matplotlib
import matplotlib.pyplot as plt


# ================================= Testing ==================================

def main(backend=None):

    if backend is not None:
        matplotlib.use(backend)

    fig1 = plt.figure()
    ax1 = fig1.add_subplot(211)
    ax2 = fig1.add_subplot(212)

    ax2.set_facecolor('k')

    ax1.plot(1, 1, 'ok')

    fig2, ax3 = plt.subplots()

    ax3.plot(1, 1, '+r')

    plt.show(block=False)

    ClickFig()


# ============================== ClickFig class ==============================


class ClickFig:
    """Mouse that activates figures and axes by hovering and clicking.

    - Left Click on figure / axes to make them the current ones in Matplotlib.
    - Right Click anywhere to deactivate the interactive mouse.

    Parameters
    ----------
    - n (int, default 1): maximum number of clicks allowed.
    - highlight (bool, default True): change ax/fig color when mouse on them.
    """

    fig_selectcolor = '#F3F8FA'
    ax_selectcolor = '#ECF4F8'

    clickfigs = []  # tracks all instances of ClickFig

    def __init__(self, n=1, highlight=True):

        self.figs = ClickFig.list_figures()

        self.n = n
        self.highlight = highlight

        if len(self.figs) > 0:

            self.active_fig = plt.gcf()
            self.active_ax = plt.gca()

            self.fig_colors, self.ax_colors = self.list_bgcolors()

            self.clicknumber = 0

            ClickFig.clickfigs.append(self)

            self.connect()

        else:
            raise FigureError("No existing Matplotlib figure!")

    @staticmethod
    def list_figures():
        fignumbers = plt.get_fignums()
        figs = [plt.figure(f) for f in fignumbers]
        return figs

    def list_bgcolors(self):
        """lists and stores all background colors of figures and axes"""
        # store background colors (facecolor) of all figures in a dict
        fig_colors = {}
        # store background colors (facecolor) of all axes in a dict
        ax_colors = {}

        for fig in self.figs:
            fig_colors[fig] = fig.get_facecolor()
            for ax in fig.axes:
                ax_colors[ax] = ax.get_facecolor()

        return fig_colors, ax_colors

    def erase(self):
        self.disconnect()

# ============================= callback methods =============================

    def on_fig_enter(self, event):
        if self.highlight is True:
            fig = event.canvas.figure
            fig.set_facecolor(ClickFig.fig_selectcolor)
            fig.canvas.draw()

    def on_fig_leave(self, event):
        if self.highlight is True:
            fig = event.canvas.figure
            original_color = self.fig_colors[fig]
            fig.set_facecolor(original_color)
            fig.canvas.draw()

    def on_ax_enter(self, event):
        if self.highlight is True:
            fig = event.canvas.figure
            ax = event.inaxes
            ax.set_facecolor(ClickFig.ax_selectcolor)
            fig.canvas.draw()

    def on_ax_leave(self, event):
        if self.highlight is True:
            fig = event.canvas.figure
            ax = event.inaxes
            original_color = self.ax_colors[ax]
            ax.set_facecolor(original_color)
            fig.canvas.draw()

    def on_press(self, event):
        fig = event.canvas.figure
        ax = event.inaxes

        if event.button == 1:  # left click : activate ax/fig

            self.clicknumber += 1

            # activate clicked figure as the current figure.
            fnum = fig.number
            plt.figure(fnum)

            # activate clicked axes as the current axes.
            if ax is None:
                print('\nActive Figure defined, no Axes on click location.')
            else:
                plt.sca(ax)
                print('\nActive Figure and Axes defined.')

        # right click or reached max click number: deactivate interactive mouse
        if event.button == 3 or self.clicknumber == self.n:

            original_color = self.fig_colors[fig]
            fig.set_facecolor(original_color)
            fig.canvas.draw()

            if ax is not None:
                original_color = self.ax_colors[ax]
                ax.set_facecolor(original_color)
                fig.canvas.draw()

            if event.button == 3:
                print('\nClickFig deactivated (left click).')
            else:
                print('\nClickFig deactivated (max number of clicks reached).')

            self.erase()

    def on_close(self, event):
        pass
        # self.erase()

# ================= connect/disconnect events and callbacks ==================

    def connect(self):
        """Connect mouse events to all existing figures"""
        for fig in self.figs:
            self.enterfig_id = fig.canvas.mpl_connect('figure_enter_event', self.on_fig_enter)
            self.leavefig_id = fig.canvas.mpl_connect('figure_leave_event', self.on_fig_leave)
            self.enterax_id = fig.canvas.mpl_connect('axes_enter_event', self.on_ax_enter)
            self.leaveax_id = fig.canvas.mpl_connect('axes_leave_event', self.on_ax_leave)
            self.pressb_id = fig.canvas.mpl_connect('button_press_event', self.on_press)
            self.close_id = fig.canvas.mpl_connect('close_event', self.on_close)

    def disconnect(self):
        """Disconnect mouse events from all figures"""
        for fig in self.figs:
            fig.canvas.mpl_disconnect(self.enterfig_id)
            fig.canvas.mpl_disconnect(self.leavefig_id)
            fig.canvas.mpl_disconnect(self.enterax_id)
            fig.canvas.mpl_disconnect(self.leaveax_id)
            fig.canvas.mpl_disconnect(self.pressb_id)
            fig.canvas.mpl_disconnect(self.close_id)


# =========================== custom Figure Error ============================


class FigureError(Exception):
    """Raised when there are no existing figures."""
    pass


# ================================ direct run ================================


if __name__ == '__main__':
    main()



"""Main module for  mainly for testing"""


import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from . import Line, Rect, Cursor

# ================================= Testing ==================================


def main(blit=True, backend=None):
    """Example of various objects in different figures and axes (for testing)."""

    if backend is not None:
        matplotlib.use(backend)

    # Draw a bunch of static objects on the figure for visual references -----

    fig1, ax1 = plt.subplots()

    tt = np.linspace(0, 4, 1000)
    xx = np.exp(-tt)
    ax1.plot(tt, xx, '-b', c='lightblue', linewidth=5)

    ax1.set_xscale('log')
    ax1.set_yscale('log')

    # Add an interactive line and an interactive rectangle -------------------

    l1 = Line(blit=False)  # blit=False should be cancelled by the last line (see below)
    r1 = Rect(linewidth=2)

    # Second figure with static objects on different subplots ----------------

    z = np.random.randn(1000)

    fig2, (ax2a, ax2b) = plt.subplots(1, 2)
    ax2a.plot(tt, xx)
    ax2b.plot(z, '-ob', c='lightsteelblue')

    ax2a.set_yscale('log')

    # Add other Lines --------------------------------------------------------

    l2 = Line(ax=ax2a, color='royalblue', ptstyle='s')
    l3 = Line(ax=ax2b)

    # this one is plot in current axes, and sets the blit behavior of all lines
    l4 = Line(blit=blit, linewidth=2, ptstyle='o')

    # Add a cursor (should follow the mouse on all figs/axes) ----------------

    c = Cursor()

    plt.show()

    return l1, r1, l2, l3, l4, c


if __name__ == '__main__':
    main()



"""InteractiveObject class. Not for direct use, just subclassing.

Line, Rect and Cursor each subclass the InteractiveObject class defined here.
"""

import matplotlib.pyplot as plt
from matplotlib.colors import is_color_like


def main():
    obj = InteractiveObject()  # For testing purposes
    return obj


class InteractiveObject:
    """Base class for moving objects on a figure. Used for subclassing only.

    See DEVELOPMENT.md for information on how to use this class.
    """

    name = 'Interactive Object'

    all_interactive_objects = []  # tracking all instances of all subclasses
    # (use the all_instances() method to get instances of a single class)
    # This list is returned by the classmethod all_objects().

    moving_objects = set()  # objects currently moving on figure. Includes
    # all subclasses, to be able to manage motion of objects of different
    # classes on the same figure.

    leader = None  # Leader object that synchronizes motion and drawing of all
    # objects when several objects are selected/moving at the same time.
    # This feature is required due to blitting rendering issues.

    initiating_motion = False  # True when the leading object had been selected

    # Attributes for fast rendering of motion with blitting.
    blit = True
    background = None

    # Define default colors of the class (potentially cycled through by some
    # methods. If user specifies a color not in the list, it is added to the
    # class colors.
    colors = ['crimson', 'dimgray', 'whitesmoke', 'dodgerblue', 'lightgreen']

    def __init__(self, fig=None, ax=None, color=None, c=None,
                 blit=True, block=False):

        self.fig = plt.gcf() if fig is None else fig
        self.ax = plt.gca() if ax is None else ax

        # Connect matplotlib event handling to callback functions
        self.connect()

        # Tracks instances of any interactive objects of any subclass.
        self.all_interactive_objects.append(self)

        self.all_artists = []  # all artists the object is made of
        self.all_pts = []  # all individual tracking points the object is made of

        # set that stores info about what artists are picked
        self.picked_artists = set()  # they can be different from the active
        # artists, because e.g. when a line moves as a whole, only the line
        # is picked, but the two edge points need to be moving/active as well.

        self.moving = False  # faster way to check moving objects than to measure the length of moving_objects
        self.press_info = {'currently_pressed': False}  # stores useful useful mouse click information

        # the last object to be instanciated dictates if blitting is true or not
        self.__class__.blit = blit
        # Reset leading artist when instanciating a new object
        self.__class__.leader = None

        # defines whether the interactive object is blocking the console or not
        self.block = block

        # color management ---------------------------------------------------

        if c is not None:
            color = c

        if color is None:
            self.color = self.__class__.colors[0]
        elif not is_color_like(color):
            print('Warning: color not recognized. Falling back to default.')
            self.color = self.__class__.colors[0]
        else:
            self.color = color
            if color not in self.__class__.colors:
                self.__class__.colors.append(color)

        # --------------------------------------------------------------------

        # Transforms functions to go from px to data coords.
        # Need to be redefined if figure is resized
        self.datatopx = self.ax.transData.transform  # transform between data coords to px coords.
        self.pxtodata = self.ax.transData.inverted().transform  # pixels to data coordinates

        # this seems to be a generic way to bring window to the front but I
        # have not checked with all backends etc, and it does not always work
        plt.get_current_fig_manager().show()

    def __repr__(self):
        object_list = self.__class__.class_objects()
        objects_on_fig = [obj for obj in object_list if obj.fig == self.fig]
        n_on_fig = len(objects_on_fig)
        try:
            n = objects_on_fig.index(self) + 1
        except ValueError:  # means object has been deleted
            n = 'deleted'
        name = self.__class__.name
        return f'{name} #{n}/{n_on_fig} in Fig. {self.fig.number}.'

    def __str__(self):
        # name = self.__class__.name
        # return f'{name} on Fig. {self.fig.number}.'
        return self.__repr__()

# ================================== methods =================================

    def update_graph(self, event):
        """Update graph with the moving artists. Called only by the leader."""

        canvas = self.fig.canvas
        ax = self.ax

        if self.__class__.blit and self.__class__.initiating_motion:
            canvas.draw()
            self.__class__.background = canvas.copy_from_bbox(ax.bbox)
            self.__class__.initiating_motion = False

        if self.__class__.blit:
            # without this line, the graph keeps all successive positions of
            # the cursor on the screen
            canvas.restore_region(self.__class__.background)

        # now the leader triggers update of all moving artists including itself
        for obj in self.__class__.moving_objects:

            # update position data of object depending on its motion mode
            obj.update_position(event)

            # Draw all artists of the object (if not, some can miss in motion)
            if self.__class__.blit:
                for artist in obj.all_artists:
                    ax.draw_artist(artist)

        # without this below, the graph is not updated
        if self.__class__.blit:
            canvas.blit(ax.bbox)
        else:
            canvas.draw()

    def initiate_motion(self, event):
        """Initiate motion and define leading artist that synchronizes plot.

        In particular, if there are several moving objects, save background
        (for blitting) only once. The line selected first becomes the leader
        for moving events, i.e. it is the one that detects mouse moving and
        triggers re-drawing of all other moving lines.
        Note : all moving lines are necesary on the same axes"""

        if self.__class__.leader is None:
            self.__class__.leader = self
            # Below is to delay background setting for blitting until all
            # artists have been defined as animated.
            # This is because the canvas.draw() and/or canvas_copy_from_bbox()
            # calls need to be made with all moving artists declared as animated
            self.__class__.initiating_motion = True
            self.fig.canvas.draw()

        self.__class__.moving_objects.add(self)
        self.moving = True
        if self.__class__.blit:
            for artist in self.all_artists:
                artist.set_animated(True)

        # find which elements need to be active/updated during mouse motion
        # and motion mode (defined in subclasses)
        self.set_active_info()
        # store location of pts and of click, and create dict of px pos data for motion
        # (defines self.press_info and self.moving_positions)
        self.set_press_info(event)

        # Transforms functions to go from px to data coords.
        # Need to be redefined if figure is resized or if zooming occurs
        self.datatopx = self.ax.transData.transform  # transform between data coords to px coords.
        self.pxtodata = self.ax.transData.inverted().transform  # pixels to data coordinates

    def reset_after_motion(self):
        """Reset attributes that should be active only during motion."""

        self.picked_artists = set()
        self.active_info = {}
        self.press_info = {'currently_pressed': False}
        self.moving_positions = {}
        self.moving = False

        if self.__class__.blit:
            for artist in self.all_artists:
                artist.set_animated(False)

        # Reset class variables that store moving information
        self.__class__.moving_objects.remove(self)
        if self is self.__class__.leader:
            self.__class__.leader = None

    def set_press_info(self, event):
        """Records information related to the mouse click, in px coordinates."""

        press_info = {'click': (event.x, event.y)}  # record click position
        moving_positions = {}

        for pt in self.all_pts:
            xpt, ypt = self.get_pt_position(pt, 'px')
            press_info[pt] = xpt, ypt  # position of all object pts during click
            moving_positions[pt] = xpt, ypt  # will be updated during motion

        self.press_info = press_info
        self.moving_positions = moving_positions

    def eraser(self, option):
        """Private erasing function that is used by erase() and delete()"""

        for artist in self.all_artists:
            artist.remove()
        self.all_artists = []

        self.fig.canvas.draw()

        # Check if object is listed as still moving, and remove it.
        moving_objects = self.__class__.moving_objects
        if self in moving_objects:
            moving_objects.remove(self)

        if option == 'erase':
            pass
        elif option == 'delete':
            self.__class__.all_interactive_objects.remove(self)
            self.disconnect()
            if self.block:
                self.fig.canvas.stop_event_loop()
        else:
            print('Warning: eraser function not called properly.')

    def erase(self):
        """Lighter than delete(), keeps object connected and referenced"""
        self.eraser('erase')

    def delete(self):
        """Hard delete of object by removing its components and references"""
        self.eraser('delete')

    def delete_others(self, *args):
        """Delete other instances of the same class (eccluding parents/children)

        Options 'all', 'fig', 'ax'. If no argument, assumes 'all'.
        """
        if len(args) == 0:
            option = 'all'
        else:
            option, *_ = args

        all_instances = self.__class__.class_objects()

        if option == 'all':
            instances = all_instances
        elif option == 'fig':
            instances = [obj for obj in all_instances if obj.fig == self.fig]
        elif option == 'ax':
            instances = [obj for obj in all_instances if obj.ax == self.ax]
        else:
            raise ValueError(f"{option} is not a valid argument for delete_others(). "
                             "Possible values: 'all', 'fig', 'ax'.")

        others = set(instances) - {self}
        for other in others:
            other.delete()

    def get_pt_position(self, pt, option='data'):
        """Gets point position as a tuple from matplotlib line object.

        Options : 'data' (axis data coords, default) or 'px' (pixel coords).
        """
        # if orig=True (default), the format is not always consistent
        xpt, ypt = pt.get_data(orig=False)
        # convert numpy array to scalar, faster than unpacking
        pos = xpt.item(), ypt.item()
        if option == 'data':
            return pos
        elif option == 'px':
            pos_px = self.datatopx(pos)
            return pos_px
        else:
            raise ValueError(f'{option} not a valid argument.')

    @classmethod
    def class_objects(cls):
        """Return all instances of a given class, excluding parent class."""
        # note : isinstance(obj, class) would also return the parent's objects
        instances = [obj for obj in cls.all_objects() if type(obj) is cls]
        return instances

    @classmethod
    def all_objects(cls):
        """Return all interactive objects, including parents and subclasses."""
        return cls.all_interactive_objects

    @classmethod
    def clear(cls):
        """Delete all interactive objects of the class and its subclasses."""
        objects = [obj for obj in cls.all_objects()]  # needed to avoid iterating over decreasing set
        for obj in objects:
            obj.delete()

# TO DEFINE IN SUBCLASSES ====================================================

    def create(self):
        """Create object based on options. Need to be defined in subclass."""
        pass

    def set_active_info(self):
        """Store information useful when in motion. Defined in subclasses."""
        active_info = {}
        return active_info

    def update_position(self, event):
        """Update object position depending on moving mode and mouse position.

        Used by update_graph.
        """
        pass

# ================= connect/disconnect events and callbacks ==================

    def connect(self):
        """connect object to figure canvas events"""
        # mouse events
        self.cidpress = self.fig.canvas.mpl_connect('button_press_event',
                                                    self.on_mouse_press)
        self.cidrelease = self.fig.canvas.mpl_connect('button_release_event',
                                                      self.on_mouse_release)
        self.cidpick = self.fig.canvas.mpl_connect('pick_event',
                                                   self.on_pick)
        self.cidmotion = self.fig.canvas.mpl_connect('motion_notify_event',
                                                     self.on_motion)
        # key events
        self.cidpressk = self.fig.canvas.mpl_connect('key_press_event',
                                                     self.on_key_press)
        self.cidreleasek = self.fig.canvas.mpl_connect('key_release_event',
                                                       self.on_key_release)
        # figure events
        self.cidfigenter = self.fig.canvas.mpl_connect('figure_enter_event',
                                                       self.on_enter_figure)
        self.cidfigleave = self.fig.canvas.mpl_connect('figure_leave_event',
                                                       self.on_leave_figure)
        self.cidaxenter = self.fig.canvas.mpl_connect('axes_enter_event',
                                                      self.on_enter_axes)
        self.cidaxleave = self.fig.canvas.mpl_connect('axes_leave_event',
                                                      self.on_leave_axes)
        self.cidclose = self.fig.canvas.mpl_connect('close_event',
                                                    self.on_close)
        self.cidresize = self.fig.canvas.mpl_connect('resize_event',
                                                     self.on_resize)

    def disconnect(self):
        """disconnect callback ids"""
        # mouse events
        self.fig.canvas.mpl_disconnect(self.cidpress)
        self.fig.canvas.mpl_disconnect(self.cidrelease)
        self.fig.canvas.mpl_disconnect(self.cidmotion)
        self.fig.canvas.mpl_disconnect(self.cidpick)
        # key events
        self.fig.canvas.mpl_disconnect(self.cidpressk)
        self.fig.canvas.mpl_disconnect(self.cidreleasek)
        # figure events
        self.fig.canvas.mpl_disconnect(self.cidfigenter)
        self.fig.canvas.mpl_disconnect(self.cidfigleave)
        self.fig.canvas.mpl_disconnect(self.cidaxenter)
        self.fig.canvas.mpl_disconnect(self.cidaxleave)
        self.fig.canvas.mpl_disconnect(self.cidclose)
        self.fig.canvas.mpl_disconnect(self.cidresize)

# ============================= callback methods =============================

    # mouse events  ----------------------------------------------------------

    def on_pick(self, event):
        """If picked, save picked objects, or delete objects if right click"""
        selected = event.artist
        if event.mouseevent.button == 3 and (selected in self.all_artists):
            self.delete()
            return
        if selected in self.all_artists:
            self.picked_artists.add(selected)

    def on_mouse_press(self, event):
        """When mouse button is pressed, initiate motion."""
        if len(self.picked_artists) == 0:
            return
        else:
            self.initiate_motion(event)

    def on_motion(self, event):
        """Leader artist triggers graph update based on mouse position"""
        if not self.moving:
            return
        # only the leader triggers moving events (others drawn in update_graph)
        if self.__class__.leader is self:
            self.update_graph(event)

    def on_mouse_release(self, event):
        """When mouse released, reset attributes to non-moving"""
        if self not in self.__class__.moving_objects:
            return
        else:
            self.reset_after_motion()

    # key events  ------------------------------------------------------------

    def on_key_press(self, event):
        print(f'Key Press: {event.key}')
        pass

    def on_key_release(self, event):
        pass

    # figure events  ---------------------------------------------------------

    def on_enter_figure(self, event):
        pass

    def on_leave_figure(self, event):
        pass

    def on_enter_axes(self, event):
        pass

    def on_leave_axes(self, event):
        pass

    def on_resize(self, event):
        """When resizing, re-adjust the conversion from pixels to data pts."""
        # Transforms functions to go from px to data coords.
        # Need to be redefined if figure is resized or if zooming occurs
        self.datatopx = self.ax.transData.transform  # transform between data coords to px coords.
        self.pxtodata = self.ax.transData.inverted().transform  # pixels to data coordinates

    def on_close(self, event):
        """Delete object if figure is closed"""
        self.delete()



if __name__ == '__main__':
    main()



"""Additional Interactive Features for Matplotlib.

Draggable shapes
- Line is a class that creates a draggable line.
- Rect is a class that creates a draggable rectangle.

Cursor and ginput
- Cursor is a class that creates a cursor that follows the mouse
- ginput is a function based on Cursor that returns data from clicks/keystrokes.

Other
- ClickFig is a class that activates figures and axes (makes them the current
ones) by mouse hovering and clicking.

"""

from .cursor import Cursor
from .cursor import ginput

from .line import Line
from .line import linput

from .rectangle import Rect
from .rectangle import rinput

from .clickfig import ClickFig

from importlib_metadata import version

__version__ = version('drapo')



"""Demo of several objects, functions and their features."""

import argparse

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from . import Line, ginput, rinput


def demo_figure():
    """Create demo fig and axes."""

    fig, (ax1, ax2) = plt.subplots(1, 2)
    fig.set_size_inches(9, 4)

    npts = 1000

    tt = np.linspace(0, 4, npts)
    rr = np.random.randn(npts)
    xx = np.exp(-tt) + 0.1 * rr

    ax1.set_yscale('log')

    ax1.plot(tt, xx, '.', c='lightsteelblue')
    ax2.plot(rr, '-o', c='lightsteelblue')

    fig.set_facecolor('gainsboro')

    return fig, (ax1, ax2)



def demo(blit=True, backend=None):
    """Example of various objects in different figures and axes (for testing)."""

    if backend is not None:
        matplotlib.use(backend)

    print('Blitting', blit, 'Backend', matplotlib.get_backend())

    # Demo of Cursor and ginput()

    fig, (ax1, ax2) = demo_figure()

    fig.suptitle('Cursor() and ginput()', fontsize=16)
    data = ginput(4, blit=blit)
    print(f'{len(data)} recorded clicks: ', data)

    fig.suptitle('Rect() and rinput()', fontsize=16)
    data = rinput(blit=blit)
    print('recorded rectangle (x, y, w, h): ', data)

    fig.suptitle('Line()', fontsize=16)
    Line(c='k', linewidth=2, ptstyle='s', ptsize=8)
    Line(c='seagreen', linewidth=4)

    # The last object imposes blitting behavior of all objects
    Line(linestyle='--', linewidth=1, ptstyle='o', blit=blit)

    plt.show()


if __name__ == '__main__':

    descr = "Run demo for drapo, with backend and blitting options."

    parser = argparse.ArgumentParser(description=descr,
                                     formatter_class=argparse.RawTextHelpFormatter)

    msg = "(str): Matplotlib backend (e.g. 'TkAgg', 'Qt5Agg', 'MacOSX', etc.)"
    parser.add_argument('-B', '--backend', type=str, help=msg)

    msg = "(True or False): use blitting for fast rendering (default True)"
    parser.add_argument('-b', '--blit', type=str, default='True', help=msg)

    args = parser.parse_args()

    blit = True if args.blit == 'True' else False

    demo(blit=blit, backend=args.backend)



"""Tests for the drapo module."""


from drapo.__main__ import main


import matplotlib
matplotlib.use('Qt5Agg')


def test_objects():
    """Test a bunch of interactive objects on a figure, including log plots."""
    try:
        main(blit=True, backend='Qt5Agg')
    except ValueError:
        main(blit=True, backend='TkAgg')





"""Interactive, draggable rectangle on matplotlib figure."""


# TODO -- Add option for appearance of center
# TODO -- add possibility to interactively change color


from .interactive_object import InteractiveObject


class Rect(InteractiveObject):
    """Interactive draggable rectangle in matplotlib figure/axes.

    Left click to drag rectangle, right click to remove it. Clicking can be
    done on the edges, vertices (corners), or on the center. These clicks
    trigger different modes of motion.

    Parameters
    ----------
    All parameters optional so that a rectangle can be created by `Rect()`.

    - `fig` (matplotlib figure, default: current figure, specified as None).
    - `ax` (matplotlib axes, default: current axes, specified as None).
    - `position` (4-tuple (xmin, ymin, width, height) in data coordinates;
                  default None, i.e. rectangle automatically centered in axes).
    - 'pickersize' (float, default: 5), tolerance for object picking.
    - `color` (matplotlib's color, default: None (class default value)).
    - `c` (shortcut for `color`)

    Appearance of the vertices (corners):
    - `ptstyle` (matplotlib's marker, default: dot '.').
    - `ptsize` (float, default: 5). Marker size.

    Appearance of the edges (lines):
    - `linestyle` (matplotlib's linestyle, default: continuous '-').
    - `linewidth` (float, default: 1). Line width.

    Other
    - `blit` (bool, default True). If True, blitting is used for fast rendering
    - `block`(bool, default False). If True, object blocks the console
    - `timeout` (float, default 0, i.e. infinite) timeout for blocking.
    """

    name = "Draggable Rectangle"

    def __init__(self, fig=None, ax=None, position=None, pickersize=5, c=None,
                 color=None, ptstyle='.', ptsize=8, linestyle='-', linewidth=1,
                 blit=True, block=False, timeout=0):

        super().__init__(fig=fig, ax=ax, color=color, c=c,
                         blit=blit, block=block)

        xlim, ylim = self.ax.get_xlim(), self.ax.get_ylim()

        self.create(pickersize, position,
                    ptstyle, ptsize, linestyle, linewidth)

        # to prevent any shift in axes limits when instanciating line. The
        self.ax.set_xlim(xlim)
        self.ax.set_ylim(ylim)

        self.fig.canvas.draw()  # Useful?

        if self.block:
            self.fig.canvas.start_event_loop(timeout=timeout)

    def create(self, pickersize, position,
               ptstyle, ptsize, linestyle, linewidth):

        positions = self.set_initial_position(position)
        corner_positions = positions[:-1]  # the last one is the center

        # Create center of rectangle -----------------------------------------
        x_center, y_center = positions[-1]
        center, = self.ax.plot(x_center, y_center, marker='+', c=self.color,
                               markersize=ptsize)

        # Create all vertices (corners) of the rectangle ---------------------
        corners = []
        for pos in corner_positions:
            pt, = self.ax.plot(*pos, marker=ptstyle, c=self.color,
                               markersize=ptsize)
            corners.append(pt)

        # Create all lines (edges) of the rectangle --------------------------
        lines = []
        for i, pos in enumerate(corner_positions):
            x1, y1 = corner_positions[i - 1]
            x2, y2 = pos
            line, = self.ax.plot([x1, x2], [y1, y2], c=self.color,
                                 linestyle=linestyle, linewidth=linewidth)
            lines.append(line)

        # Define useful collections of lines / pts ---------------------------

        self.corners = corners
        self.edges = lines
        self.center = center

        self.all_artists = (*corners, *lines, center)
        self.all_pts = (*corners, center)

        # make all components of the objects pickable ------------------------
        for artist in self.all_artists:
            artist.set_picker(True)

        # Adjust pick tolerance depending on component -----------------------
        for pt in self.all_pts:
            pt.set_pickradius(pickersize + ptsize / 2)

        for line in lines:
            line.set_pickradius(pickersize)

    def set_initial_position(self, position):
        """Set position of new line, avoiding existing lines if necessary.

        Note: this is how the corners and lines are listed:

                          line 3
        corner 3   ____________________  corner 2
                   |                   |
                   |                   |
        line 0     |                   |  line 2
                   |                   |
                   |                   |
        corner 0   --------------------- corner 1
                           line 1
        """
        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()

        # Move into px coordinates to avoid problems with nonlinear axes
        xmin, ymin = self.datatopx((xmin, ymin))
        xmax, ymax = self.datatopx((xmax, ymax))

        if position is None:
            w, h = .5, .5  # relative initial width/height of rectangle in axes
            x_left = (1 - w) / 2 * (xmax - xmin) + xmin
            x_right = (1 + w) / 2 * (xmax - xmin) + xmin
            y_low = (1 - h) / 2 * (ymax - ymin) + ymin
            y_high = (1 + h) / 2 * (ymax - ymin) + ymin
        else:
            x0, y0, width, height = position
            x_left, y_low = self.datatopx((x0, y0))
            x_right, y_high = self.datatopx((x0 + width, y0 + height))

        x_center = (x_left + x_right) / 2
        y_center = (y_low + y_high) / 2

        pos1 = self.pxtodata((x_left, y_low))
        pos2 = self.pxtodata((x_right, y_low))
        pos3 = self.pxtodata((x_right, y_high))
        pos4 = self.pxtodata((x_left, y_high))
        posc = self.pxtodata((x_center, y_center))

        return pos1, pos2, pos3, pos4, posc

    @staticmethod
    def corners_to_edge(icorner1, icorner2):
        """Find the edge associated with two corners"""
        icorners = set([icorner1, icorner2])
        nozero = icorners - {0}
        if len(nozero) == 1:
            val, = nozero
            output = 1 if val == 1 else 0
        else:
            output = max(nozero)
        return output

    @staticmethod
    def edges_to_corner(iline1, iline2):
        """Find the corner associated with two edge lines"""
        ilines = set([iline1, iline2])
        nozero = ilines - {0}
        if len(nozero) == 1:
            val, = nozero
            output = 3 if val == 3 else 0
        else:
            output = min(nozero)
        return output

    def get_position(self):
        """Get position (xmin, ymin, width, height) in data coordinates."""
        xs, ys = [], []
        corners = self.all_artists[:4]
        for pt in corners:
            x, y = self.get_pt_position(pt, 'data')
            xs.append(x)
            ys.append(y)
        try:
            xmin, xmax, ymin, ymax = min(xs), max(xs), min(ys), max(ys)
        except ValueError:
            raise ValueError('Impossible to get position of Rect, probably '
                             'because it has been deleted from the figure or '
                             'because figure itself has been closed.')
        w, h = xmax - xmin, ymax - ymin
        return xmin, ymin, w, h

    def _set_center_mode(self):
        """Given a selected line index, determine mode and active lines/pts."""
        self.active_info = {'pts': self.all_pts,
                            'lines': self.edges,
                            'mode': 'center'}

    def _set_edge_mode(self, iline):
        """Given a selected line index, determine mode and active lines/pts."""

        line = self.edges[iline]

        # Distinguish whether line picked is vertical or horizontal
        mode = 'horz_edge' if iline % 2 else 'vert_edge'

        # neighbors of that line
        inext = (iline + 1) % 4
        iprev = (iline - 1) % 4
        active_lines = [line, self.edges[inext], self.edges[iprev]]

        # points connected to that line
        active_pts = [self.corners[iprev], self.corners[iline], self.center]

        self.active_info = {'pts': active_pts,
                            'lines': active_lines,
                            'mode': mode}

    def _set_corner_mode(self, icorner):
        """Given a selected corner, figure out active lines/pts."""

        iprev = (icorner - 1) % 4
        inext = (icorner + 1) % 4

        active_pts = [self.corners[icorner], self.corners[iprev],
                      self.corners[inext], self.center]
        active_lines = self.edges  # all lines need to be updated in corner motion

        self.active_info = {'pts': active_pts,
                            'lines': active_lines,
                            'mode': icorner}

    def set_active_info(self):
        """Set active/inactive points during motion and detect motion mode."""

        if self.center in self.picked_artists:
            # center has been picked --> solid body motion
            self._set_center_mode()

        else:
            # Need to figure out which lines and corners have been picked
            picked_corners = set.intersection(self.picked_artists, self.corners)
            ncorners = len(picked_corners)

            if ncorners > 2:
                # If for some reason (e.g. overlap) this happens, also do
                # solid-body motion
                self._set_center_mode()

            elif ncorners == 2:
                # This is also an overlapping situation where two corners
                # correponding to an edge are selected --> edge motion
                icorners = [self.corners.index(pt) for pt in picked_corners]
                iline = self.corners_to_edge(*icorners)
                self._set_edge_mode(iline)

            elif ncorners == 1:
                # Just one corner picked: corner motion, move corresponding two edges
                corner, = picked_corners
                icorner = self.corners.index(corner)
                self._set_corner_mode(icorner)

            else:
                # No corner picked, but there might have been an edge line or more
                picked_lines = set.intersection(self.picked_artists, self.edges)
                nlines = len(picked_lines)

                if nlines == 1:
                    # Just one line: edge motion with that line
                    line, = picked_lines
                    i = self.edges.index(line)
                    self._set_edge_mode(i)

                elif nlines == 2:
                    # Two lines (and for some reason no corner): corner mode
                    ilines = [self.edges.index(line) for line in picked_lines]
                    icorner = self.edges_to_corner(*ilines)
                    self._set_corner_mode(icorner)

                elif nlines > 2:
                    # More than two lines: if this happens, whole object is kind of selected
                    self._set_center_mode()

                else:
                    # Normally, this case should never happen.
                    print('Warning: Rect.set_active_info() called while no artists '
                          'picked. Please report bug.')

    def update_position(self, event):
        """Update object position depending on moving mode and mouse position."""

        x, y = event.x, event.y  # pixel positions

        active_pts = self.active_info['pts']
        active_lines = self.active_info['lines']
        mode = self.active_info['mode']

        if mode in ['horz_edge', 'vert_edge', 'center']:

            # get where click was initially made and calculate motion
            x0, y0 = self.press_info['click']
            dx = x - x0 if mode != 'horz_edge' else 0
            dy = y - y0 if mode != 'vert_edge' else 0

            for pt in active_pts:
                x0_pt, y0_pt = self.press_info[pt]
                if pt != self.center:
                    x_new, y_new = x0_pt + dx, y0_pt + dy
                else:   # center point
                    norm = 1 if mode == 'center' else 0.5
                    x_new = x0_pt + dx * norm
                    y_new = y0_pt + dy * norm
                self.moving_positions[pt] = x_new, y_new

        elif mode in range(4):  # corner motion

            i = mode
            picked_corner = self.corners[i]
            next_corner = self.corners[(i + 1) % 4]
            prev_corner = self.corners[(i - 1) % 4]
            opposite_corner = self.corners[(i + 2) % 4]

            self.moving_positions[picked_corner] = x, y
            xprev, yprev = self.moving_positions[prev_corner]
            xnext, ynext = self.moving_positions[next_corner]

            if i % 2:  # bottom right corner or top left
                self.moving_positions[prev_corner] = xprev, y
                self.moving_positions[next_corner] = x, ynext
            else:  # bottom left or top right corner
                self.moving_positions[prev_corner] = x, yprev
                self.moving_positions[next_corner] = xnext, y

            # Calculate center pos from picked pt and diagonally opposed one
            x1, y1 = self.moving_positions[picked_corner]
            x2, y2 = self.moving_positions[opposite_corner]
            self.moving_positions[self.center] = (x1 + x2) / 2, (y1 + y2) / 2

        # now apply the changes to the graph ---------------------------------
        for pt in active_pts:
            xnew, ynew = self.pxtodata(self.moving_positions[pt])
            pt.set_data(xnew, ynew)

        for line in active_lines:
            i = self.edges.index(line)
            x1, y1 = self.pxtodata(self.moving_positions[self.corners[i - 1]])
            x2, y2 = self.pxtodata(self.moving_positions[self.corners[i]])
            line.set_data([x1, x2], [y1, y2])

# ============================= callback methods =============================

    def on_key_press(self, event):
        if event.key == 'enter':
            print('Rectangle position recorded. Rectangle deleted.')
            self.recorded_position = self.get_position()
            self.delete()


# ======================== function rinput ===================================


def rinput(**kwargs):
    """Select area of figure with interactive rectangle (enter to validate).

    Parameters
    ----------
    Same as drapo.Rect(), but block remains always True by default

    Returns
    -------
    4-tuple (xmin, ymin, width, height) of data coordinates.
    """
    if 'block' in kwargs:
        kwargs.pop('block')

    r = Rect(block=True, **kwargs)
    try:
        position = r.recorded_position
    except AttributeError:
        print('Warning - Rectangle deleted before data validation. Returning None.')
        position = None
    return position



"""Extensions to Matplotlib: Line class (draggable line)"""

# TODO -- add double click to "freeze" line to avoid moving it by mistake later?
# TODO -- add more keystroke controls ?
# TODO -- key press to make the line exactly vertical or horiztontal
# TODO -- live indication of the slope of the line.

import math as m

from .interactive_object import InteractiveObject


class Line(InteractiveObject):
    """Interactive draggable line in matplotlib figure/axes.

    Left click anywhere on the line to drag it, right click to remove it.

    Parameters
    ----------
    All parameters optional so that a line can simply be created by `Line()`.

    - `fig` (matplotlib figure, default: current figure, specified as None).
    - `ax` (matplotlib axes, default: current axes, specified as None).
    - 'pickersize' (float, default: 5), tolerance for line picking.
    - `color` (matplotlib's color, default: None (class default value)).
    - `c` (shortcut for `color`)

    Appearance of the edge points:
    - `ptstyle` (matplotlib's marker, default: dot '.').
    - `ptsize` (float, default: 5). Marker size.

    Appearance of the connecting line:
    - `linestyle` (matplotlib's linestyle, default: continuous '-').
    - `linewidth` (float, default: 1). Line width.

    Instantiation option:
    - `avoid_existing` (bool, default:True). Avoid overlapping existing lines
    (only avoids that edge points overlap, but lines can still cross).

    Other
    - `blit` (bool, default True). If True, blitting is used for fast rendering
    - `block`(bool, default False). If True, object blocks the console
    - `timeout` (float, default 0, i.e. infinite) timeout for blocking.
    """

    name = 'Draggable Line'

    def __init__(self, fig=None, ax=None, pickersize=5, color=None, c=None,
                 ptstyle='.', ptsize=8, linestyle='-', linewidth=1,
                 avoid_existing=True, blit=True, block=False, timeout=0):

        super().__init__(fig=fig, ax=ax, color=color, c=c,
                         blit=blit, block=block)

        xlim, ylim = self.ax.get_xlim(), self.ax.get_ylim()

        self.create(pickersize, self.color, ptstyle, ptsize,
                    linestyle, linewidth, avoid_existing)

        # to prevent any shift in axes limits when instanciating line. The
        self.ax.set_xlim(xlim)
        self.ax.set_ylim(ylim)

        self.fig.canvas.draw()  # Useful?

        if self.block:
            self.fig.canvas.start_event_loop(timeout=timeout)

    # ========================== main line methods ===========================

    # these methods are specific to the Line class and are called by the
    # more generic callback functions defined in InteractiveObject or below

    def create(self, pickersize, color, ptstyle, ptsize, linestyle,
               linewidth, avoid_existing):
        """Create the line and its components (pt1, pt2, link)"""

        # set position of line on screen so that it does not overlap others --
        pos = self.__class__.set_initial_position(self, pickersize, avoid_existing)
        (x1, y1), (x2, y2) = pos

        # create edge points -------------------------------------------------
        pt1, = self.ax.plot(x1, y1, marker=ptstyle, color=color, markersize=ptsize)
        pt2, = self.ax.plot(x2, y2, marker=ptstyle, color=color, markersize=ptsize)

        # create connecting line (link) ---------------------------------------
        x1, y1 = self.get_pt_position(pt1, 'data')
        x2, y2 = self.get_pt_position(pt2, 'data')

        link, = self.ax.plot([x1, x2], [y1, y2], c=color,
                             linestyle=linestyle, linewidth=linewidth)

        # assemble lines and pts into "all" ----------------------------------
        self.link = link
        self.all_artists = pt1, pt2, link
        self.all_pts = pt1, pt2

        # make all components of the objects pickable ------------------------
        for artist in self.all_artists:
            artist.set_picker(True)

        # Adjust pick tolerance depending on component -----------------------
        link.set_pickradius(pickersize)
        for pt in self.all_pts:
            pt.set_pickradius(pickersize + ptsize / 2)

    def set_initial_position(self, pickersize, avoid=True):
        """Set position of new line, avoiding existing lines if necessary."""

        a1, b1, a2, b2 = .2, .2, .8, .8  # relative initial position in axes

        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()
        # Move into px coordinates to avoid problems with nonlinear axes
        xmin, ymin = self.datatopx((xmin, ymin))
        xmax, ymax = self.datatopx((xmax, ymax))

        # default positions
        x1, y1 = (1 - a1) * xmin + a1 * xmax, (1 - b1) * ymin + b1 * ymax
        x2, y2 = (1 - a2) * xmin + a2 * xmax, (1 - b2) * ymin + b2 * ymax

        if avoid:
            mindist = 3 * pickersize  # min distance between pts to avoid overlapping
            dragonax = []  # list of coords (px) of existing lines in the current axes

            otherlines = set(self.class_objects()) - set([self])
            for line in otherlines:
                # if on same axis, record coords in a list to check overlap later
                if line.ax is self.ax:
                    pt1, pt2 = line.all_pts
                    x1b, y1b = self.get_pt_position(pt1, 'px')
                    x2b, y2b = self.get_pt_position(pt2, 'px')
                    dragonax.append((x1b, y1b))
                    dragonax.append((x2b, y2b))

            while True:
                for (xb, yb) in dragonax:
                    d1 = m.hypot(x1 - xb, y1 - yb)
                    d2 = m.hypot(x2 - xb, y2 - yb)
                    dmin = min(d1, d2)
                    if dmin < mindist:  # some of the points are too close
                        x1 += -mindist  # shift everything in a parallel manner
                        y1 += +mindist
                        x2 += -mindist
                        y2 += +mindist
                        break
                else:  # if the inner for loop finishes, it means all points are ok
                    break  # --> exit the outside while loop

        return self.pxtodata((x1, y1)), self.pxtodata((x2, y2))

    def set_active_info(self):
        """Set active/inactive points during motion and detect motion mode."""

        if len(self.picked_artists) == 1 and self.link in self.picked_artists:
            # only the line is selected --> move as a whole, make both self.pts active
            active_pts = set(self.all_pts)

        else:
            # in all other cases, the points that need to be active are
            # all the ones picked. If one pt is picked, move only this point
            # and the line (edge mode). If two pts are picked (e.g. because
            # they overlap), move the whole thing as a whole again. If nothing
            # is picked, nothing happens (should not happen because the current
            # method is only called for active objects)
            active_pts = set(self.picked_artists) - {self.link}

        npts = len(active_pts)

        if npts > 1:
            self.active_info = {'mode': 'whole', 'pts': active_pts}
        elif npts == 1:
            self.active_info = {'mode': 'edge', 'pts': active_pts}
        else:
            # Again, this should not happen
            print('Warning: Line.set_active_info() called while no artists '
                  'picked. Please report bug.')

    def update_position(self, event):
        """Update object position depending on moving mode and mouse position."""

        x, y = event.x, event.y  # pixel coordinates

        mode = self.active_info['mode']
        active_pts = self.active_info['pts']

        # EDGE mode: move just one point, the other one stays fixed ----------
        if mode == 'edge':
            pt, = active_pts  # should be the only pt in active_pts
            self.moving_positions[pt] = x, y

        # WHOLE mode: move the line as a whole in a parallel fashion ---------
        else:
            # get where click was initially made and calculate motion
            x0, y0 = self.press_info['click']
            dx, dy = x - x0, y - y0

            for pt in active_pts:
                x0_pt, y0_pt = self.press_info[pt]
                self.moving_positions[pt] = x0_pt + dx, y0_pt + dy

        # now apply the changes to the graph
        for pt in active_pts:
            xnew, ynew = self.pxtodata(self.moving_positions[pt])
            pt.set_data(xnew, ynew)

        pt1, pt2, link = self.all_artists
        x1, y1 = self.pxtodata(self.moving_positions[pt1])
        x2, y2 = self.pxtodata(self.moving_positions[pt2])
        link.set_data([x1, x2], [y1, y2])

    def get_position(self):
        """Get position ((x1, y1), (x2, y2)) in data coordinates."""
        positions = []

        for pt in self.all_pts:
            position = self.get_pt_position(pt, 'data')
            positions.append(position)

        return tuple(positions)

# ============================= callback methods =============================

    def on_key_press(self, event):
        if event.key == 'enter':
            print('Line position recorded. Line deleted.')
            self.recorded_position = self.get_position()
            self.delete()


# ======================== function rinput ===================================


def linput(**kwargs):
    """Select position information with interactive Line (enter to validate).

    Parameters
    ----------
    Same as drapo.Line(), but block remains always True by default

    Returns
    -------
    Tuple of tuples ((x1, y1), (x2, y2)) of data coordinates (unordered)
    """
    if 'block' in kwargs:
        kwargs.pop('block')

    line = Line(block=True, **kwargs)
    try:
        position = line.recorded_position
    except AttributeError:
        print('Warning - Line deleted before data validation. Returning None.')
        position = None
    return position
